from ppms import ppms
from . import transform_data
from . import default_values

from ..mspdi_2003_classes import *
from .. import format
from .. import fetch_project_data as fetch_data
from .. import info_handler
from .. import messages as msg

class PLANTACalendar(object):
    def __init__(self, cal_id=None):
        self.cals = dict()
        self._add_cal(cal_id)
        #this is used to estimate whether a day is a working day or not
        # whenever the calendar does not contain a period for that date
        self.week_days = get_default_days(
            fetch_data.get_calendar_rec(cal_id))
        self.cal_period_exceeded = False

    def _build_period_dict(self, periods):
        cal_dict = dict()
        for period in periods:
            cal_dict[period.period.get_value()] = period.work.get_value()
        return cal_dict

    def _has_cal(self, cal_id):
        return self._get_cal(cal_id) is not None
    
    def _get_cal(self, cal_id):
        if cal_id is None and hasattr(self, 'basecal_id'):
            return self.cals[self.basecal_id]
        elif cal_id is not None and cal_id in self.cals:
            return self.cals[cal_id]
        else:
            return None

    def _add_cal(self, cal_id):
        if self._has_cal(cal_id):
            return

        cal_rec = fetch_data.get_calendar_rec(cal_id)
        if cal_rec is not None:
            if cal_id is None:
                cal_id = cal_rec.calendar_id.get_value()
                self.basecal_id = cal_id
            periods = fetch_data.get_periods(cal_rec)
            self.cals[cal_id] = self._build_period_dict(periods)

    def get_working_days(self, cal_id, start_date, end_date):
        self._add_cal(cal_id)
        cal = self._get_cal(cal_id)
        if cal is None: return 0
        
        working_days = 0        
        for day in xrange(start_date, end_date + 1):
            if cal[day]:
                working_days += 1
        return working_days

    def get_workday_by_period(self, cal_id, start_date, working_days):
        self._add_cal(cal_id)
        cal = self._get_cal(cal_id)
        if cal is None or start_date is None: return 0

        step = 1 if working_days >= 0 else -1
        end_date = start_date
        while int(working_days) > 0:
            end_date += step

            if not end_date in cal:
                if self.week_days[date_to_week_day(end_date)]:
                    working_days -= step
                    
            elif cal[end_date]:
                working_days -= step 
        return end_date

def build_time(hour, part_hour=None):
    if part_hour is None:
        return '{0:0>2}:00:00'.format(int(hour))
    else:
        minute = int(part_hour * (60/100.0))
        return '{0:0>2}:{1:0>2}:00'.format(hour, minute)

def get_working_times(work_per_day, start_time=8, with_break=True):
    
    working_times = list()
    
    if not with_break:
        return [(build_time(start_time),
            build_time(start_time + int(work_per_day), 
            round(work_per_day - int(work_per_day), 2) * 100))]
    
    if work_per_day > 4:
        working_times.append((build_time(start_time), 
            build_time(start_time + 4)))
        working_times += get_working_times(work_per_day - 4, start_time + 5)
    else:
        working_times.append((build_time(start_time),
            build_time(start_time + int(work_per_day), 
            round(work_per_day - int(work_per_day), 2) * 100)))
    
    return working_times

def generate_WorkingTimes(work_per_day):
    wtimes = WorkingTimes()
    if work_per_day > 12:
        with_break = False
        start_time = 0
    else:
        with_break = True
        start_time = 8
        
    for start, finish in get_working_times(work_per_day, start_time, with_break):
        wtimes.add_WorkingTime(WorkingTime(start, finish))
    
    return wtimes

def generate_WorkingDay(day_of_week, work_per_day):
    working_day = WeekDay(DayType=day_of_week, DayWorking='1',
        WorkingTimes=generate_WorkingTimes(work_per_day))
    return working_day

def generate_NonWorkingDay(day_of_week):
    nw_day = WeekDay(DayType=day_of_week, DayWorking='0')
    return nw_day

def generate_standard_week(week_day_dict, work_per_day):
    weekdays = list()

    for week_day in week_day_dict:
        if week_day_dict[week_day]:
            weekdays.append(generate_WorkingDay(week_day, work_per_day))
        else:
            weekdays.append(generate_NonWorkingDay(week_day))
        
    return weekdays

def build_exception_day(period_rec, work_per_day):
    date_val = period_rec.period.get_value()
    day_time_start = format.day_begin_to_msp_date(date_val)
    day_time_finish = format.day_end_to_msp_date(date_val)
    
    working = '1' if period_rec.work.get_value() == 1 else '0'
    
    exception_day = WeekDay(DayType=0, DayWorking=working,
        TimePeriod=TimePeriod(day_time_start, day_time_finish))
    
    if working == '1' and work_per_day is not None:
        #build the new DayType
        exception_day.set_WorkingTimes(generate_WorkingTimes(work_per_day))
        
    return exception_day
    
def date_to_week_day(date_value):
    import time
    #date_value contains an integer for days since 1970
    week_day = time.gmtime(date_value * 3600 * 24).tm_wday
    week_day = (week_day + 2) if week_day in range(0, 6) else 1
    assert(week_day in range(1, 8))
    return week_day
    
def period_is_exception_day(period_rec, week_day_dict):
    date_val = period_rec.period.get_value()
    work_val = True if period_rec.work.get_value() == 1 else False
    
    day_time_start = format.day_begin_to_msp_date(date_val)
    
    return week_day_dict[date_to_week_day(date_val)] != work_val

def get_default_days(cal_rec):
    # please note: as opposed to it's documentation (in schema definition, weekday begins with sunday!
    dummy_result = {1:False, 2:True, 3:True, 4:True, 5:True, 6:True, 7:False}
    standard_week = {}
    di_names = ['sunday', 'monday', 'tuesday', 'wednesday', 'thursday',
                'friday', 'saturday']
    for day_type in range(1, 8):
        working_di = cal_rec.get_di(di_names[day_type - 1])
        if working_di is None:
            info_handler.warn(msg.get_const('standard_week_dis'))
            return dummy_result
        else:
            standard_week[day_type] = \
                True if working_di.get_value() == 1 else False

    if sum(standard_week.values()) == 0:
        info_handler.warn(msg.get_const('standard_week_not_set').format(cal_rec.calendar_id.get_value()))

    return standard_week

def period_in_boundaries(period_rec, project_boundaries):
    start_date, end_date = project_boundaries
    period_date = period_rec.period.get_value()
    return start_date <= period_date <= end_date        
    
def transform_periods(cal_rec, work_per_day, project_boundaries):
    week_day_dict = get_default_days(cal_rec)
    week_days = generate_standard_week(week_day_dict, work_per_day) #no need for the periods in here
    
    periods = fetch_data.get_periods(cal_rec)
    periods_obj = WeekDays()
    exception_days = list()
    for period in periods:
        if period_in_boundaries(period, project_boundaries):
            if period_is_exception_day(period, week_day_dict):
                exception_days.append(build_exception_day(period, 
                    work_per_day))
            
    periods_obj.set_WeekDay(week_days + exception_days)
    return periods_obj

def transform_res_cal_rec(base_cal_uid, cal_uid, cal_name):
    cal_object = Calendar()
    cal_object.set_UID(cal_uid)
    cal_object.set_IsBaseCalendar('0')
    cal_object.set_BaseCalendarUID(base_cal_uid)
    cal_object.set_Name(cal_name)
    
    return cal_object

def transform_base_cal_rec(cal_rec, work_per_day, uid, project_boundaries):
    cal_object = Calendar()
    cal_object.set_UID(str(uid))
    cal_object.set_IsBaseCalendar('1')
    cal_object.set_BaseCalendarUID('-1')
    cal_object.set_Name(cal_rec.calendar_id.get_value() + 
        '_{0}h'.format(work_per_day))
    cal_object.p_worktime = work_per_day
    cal_object.set_WeekDays(transform_periods(cal_rec, work_per_day,
        project_boundaries))
    return cal_object

#these may be multiple ones, as one base calendar for one work time is generated
def get_base_cals_by_name(calendar_objs, name):
    return [cal for cal in calendar_objs if cal.get_Name().rsplit('_',1)[0]\
        == name and cal.get_IsBaseCalendar() == '1']

# the cal list contains only calendars from same PLANTA origin (e.g. 'BK'), but
# with different work times
def get_cal_by_work_time(base_cals, work_time):
    for cal in base_cals:
        if work_time == cal.p_worktime:
            return cal
    
def base_cal_get_uid(cal):
    if cal is None:
        return -1
    
    return int(cal.get_UID())

def base_cal_copy(base_cal, work_per_day, uid):
    import copy
    cal_object = Calendar(UID=str(uid), IsBaseCalendar='1', BaseCalendarUID='-1',
        Name=base_cal.get_Name().rsplit('_',1)[0] + '_{0}h'.format(work_per_day))

    cal_object.p_worktime = work_per_day

    week_days = copy.deepcopy(base_cal.get_WeekDays())
    for week_day in week_days.get_WeekDay():
        if week_day.get_DayWorking() == '1':
            week_day.set_WorkingTimes(generate_WorkingTimes(work_per_day))
    cal_object.set_WeekDays(week_days)
    return cal_object
    
def transform_base_calendar(cal_objs, work_per_day, project_boundaries, cal_id=None):
    cal_obj = None
    base_cal_objs = get_base_cals_by_name(cal_objs, cal_id)
    timed_cal_obj = get_cal_by_work_time(base_cal_objs, work_per_day)
    
    if timed_cal_obj is not None:
        return timed_cal_obj

    msp_cal_uid = len(cal_objs) + 1
    if len(base_cal_objs) > 0:
        cal_obj = base_cal_copy(base_cal_objs[0], work_per_day, msp_cal_uid)

    else:
        cal_rec = fetch_data.get_calendar_rec(cal_id)
        if cal_rec is None:
            if cal_id is None:
                info_handler.warn(msg.get_const('syscalrec_fetch_failed'))
            else:
                info_handler.warn(msg.get_const('calrec_fetch_failed').format(cal_id))
            return None
        
        #check calendar against project boundaries
        cal_start = cal_rec.start_period.get_value()
        cal_end = cal_rec.end_period.get_value()
        proj_start, proj_end = project_boundaries
        if proj_start < cal_start or proj_end > cal_end:
            info_handler.warn(msg.get_const('cal_not_covering_project').format(cal_rec.calendar_id.get_value()))
            
            for workday in get_default_days(cal_rec).values():
                if workday:
                    break
            else:
                # We could never calculate a proper end date with these conditions
                raise ValueError('Calendar doesn\'t cover project + no standard workdays defined')
                
        cal_obj = transform_base_cal_rec(cal_rec, work_per_day, msp_cal_uid,
        project_boundaries)

    cal_objs.append(cal_obj)
    assert(cal_obj is not None)

    return cal_obj

def transform_calendars(project_boundaries):
    calendars = Calendars()
    
    #set the project boundaries, these will control the period the calendar is generated from
    Calendars.project_boundaries = project_boundaries
    
    # we'll add only the base calendar for now, resource calendars follow
    # during resource transformation
    info_handler.info(msg.get_const('cal_transformation'))
    system_base_cal = transform_base_calendar([], 8, project_boundaries)
    
    if system_base_cal is None:
        info_handler.error(msg.get_const('syscal_fetch_failed'))
        return None
    
    calendars.add_Calendar(system_base_cal)

    return calendars
    
def get_calendars(project_obj):
    #get_Calendar returns a list attribute containing the calendar objects
    return project_obj.get_Calendars().get_Calendar()

def get_periods_dict(period_recs, project_boundaries=None):
    #filter the records by the project boundaries if possible
    filtered_period_recs = period_recs
    if project_boundaries is not None:
        filtered_period_recs = filter(period_in_boundaries, period_recs)
    
    period_dict = dict()
    for period in filtered_period_recs:
        period_dict[period.period.get_value()] = period
    return period_dict

    
def build_res_cal(res_rec, base_cal_rec, base_cal_uid, cal_objs, work_per_day,
    project_boundaries):
    
    res_cal_obj = transform_res_cal_rec(base_cal_uid,
        len(cal_objs) + 1, res_rec.res_id.get_value())
    week_days = WeekDays()
    res_cal_obj.set_WeekDays(week_days) 

    res_periods = get_periods_dict(fetch_data.get_resource_periods(res_rec))
    base_cal_periods = get_periods_dict(fetch_data.get_periods(base_cal_rec))

    for period in res_periods:
        if period in base_cal_periods:
            if res_periods[period].work.get_value() != \
                base_cal_periods[period].work.get_value():
                
                week_days.add_WeekDay(build_exception_day(
                    res_periods[period], work_per_day))
        else:
            week_days.add_WeekDay(build_exception_day(
                res_periods[period], work_per_day))
    
    # Every Calendar needs a standard week (implicitely a Calendar contains
    # only Non-Working days...) Microsoft, what is that Base Calendar for?
    week_day_list = week_days.get_WeekDay()
    standard_week_days = generate_standard_week(get_default_days(
            base_cal_rec), work_per_day)
    
    if week_day_list is not None and type(week_day_list) == type(list()):
        week_day_list += standard_week_days
    else:
        week_day_list = standard_week_days
    
    week_days.set_WeekDay(week_day_list) 
    
    cal_objs.append(res_cal_obj)
    return str(len(cal_objs))
    
def build_resource_cal(res_rec, proj_obj):
    base_cal_id = res_rec.calendar.get_value()
    calendars_obj = proj_obj.get_Calendars()
    cal_objs = calendars_obj.get_Calendar()
    
    work_per_day = res_rec.avail_cap_base_per.get_value()
    
    if work_per_day > 24 or work_per_day == 0:
        #these resource calendars can't be transformed properly!
        info_handler.warn(msg.get_const('resource_wt_too_big').format(
            res_rec.res_id.get_value()))
        work_per_day = 8 # will use standard base calendar
    
    base_cal_obj = transform_base_calendar(cal_objs, work_per_day, 
        calendars_obj.project_boundaries, base_cal_id)
    base_cal_rec = fetch_data.get_calendar_rec(base_cal_id)
    if base_cal_obj is None:
        info_handler.warn(msg.get_const('no_base_cal').format(
            res_rec.res_id.get_value()))
        return None
    
    return build_res_cal(res_rec, base_cal_rec, base_cal_obj.get_UID(), 
        cal_objs, work_per_day, calendars_obj.project_boundaries)

def build_cal(cal_id, existing_cals, work_per_day, boundaries):
    cal_rec = fetch_data.get_calendar_rec(cal_id)

    #generates the whole data (standard week + exception days)
    week_days = transform_periods(cal_rec, work_per_day, boundaries)
    uid = str(max([int(cal.get_UID()) for cal in existing_cals]) + 1)
 
    new_cal = Calendar(UID=uid, Name=cal_id, IsBaseCalendar='1', BaseCalendarUID='-1')
    new_cal.set_WeekDays(week_days)
    return new_cal

def get_cal(cal_id, calendars, work_per_day, boundaries):
    cal_objs = get_base_cals_by_name(calendars.get_Calendar(), cal_id)
    assert(len(cal_objs) <= 1)
    if len(cal_objs) > 1:
        timed_cal_obj = get_cal_by_work_time(cal_objs, work_per_day)
        return timed_cal_obj.get_UID()
    elif len(cal_objs) == 1:
        return cal_objs[0].get_UID()
    else:
        new_cal = build_cal(cal_id, calendars.get_Calendar(), work_per_day, boundaries)
        calendars.add_Calendar(new_cal)
        return new_cal.get_UID()

def get_extra_cal(cal_id, proj_obj):
    uid = get_cal(cal_id, proj_obj.get_Calendars(),
        8, proj_obj.get_Calendars().project_boundaries)
    return uid

