from ppms import ppms
from . import transform_data
from . import default_values
from . import transform_resource_structure as transform_resource
from . import transform_calendar
from .transform_calendar import PLANTACalendar 

from ..mspdi_2003_classes import *
from .. import format
from .. import fetch_project_data as fetch_data
from .. import info_handler
from .. import messages as msg
from .. import task_hash

def transform_task(task_rec):
    info_handler.info(msg.get_const('task_transformation').format(
        task_rec.task_id.get_value()))
    
    task_obj = Task()
    
    # Set Task default values
    default_values.add_task_defaults(task_obj)
    
    transform_data.transform_task_data(task_rec, task_obj)
    return task_obj

def transform_link(link_rec, pred_uid):
    link_obj = PredecessorLink(PredecessorUID = pred_uid)
    transform_data.transform_link_data(link_rec, link_obj)
    return link_obj

def find_task_rec_by_planta_id(task_records, id):
    for task in task_records:
        if task.task_id.get_value() == id:
            return task
    return None

def calc_correction_lag(task_rec, succ_rec, link_type, lag, cal):
    # PLANTA shows day based planning but calculates implicitely
    # with a more accurate time. This causes problems in the export,
    # as we can't get access to the accurate starting times e.g. when a
    # successor task begins at the same day as its predecessor ends
    # PLANTA SS
    succ_start = succ_rec.calc_start.get_value()
    succ_end = succ_rec.calc_end.get_value()
    task_start = task_rec.calc_start.get_value()
    task_end = task_rec.calc_end.get_value()

    if link_type == 1 and \
        succ_start < cal.get_workday_by_period(None, task_start, lag):
        return lag - (cal.get_workday_by_period(None, task_start, lag) - succ_start)
    # PLANTA SE
    elif link_type == 2 and \
        succ_end < cal.get_workday_by_period(None, task_start, lag):
        return lag - (cal.get_workday_by_period(None, task_start, lag) - succ_end)
    # PLANTA ES
    elif link_type == 3 and \
        succ_start <= cal.get_workday_by_period(None, task_end, lag+1):
        return lag - (cal.get_workday_by_period(None, task_end, lag+1) - succ_start + 1)
    #PLANTA EE
    elif link_type == 4 and \
        succ_end < cal.get_workday_by_period(None, task_end, lag):
        return lag - (cal.get_workday_by_period(None, task_end, lag) - succ_end)
    else:
        return lag

def calculate_link_lag(link, task_rec):
    if link.time_interval.get_value() > 0:
        lag_value = link.time_interval.get_value()

        if link.link_type_percent.get_value() > 0:
            task_duration = task_rec.duration_plan.get_value()
            
            return (lag_value / 100.0) * task_duration
        else:
            return lag_value
    else: return 0

def add_link_lag(link_obj, lag):
    if lag != 0.0:
        link_obj.set_LinkLag(format.link_lag_to_msp_link_lag(lag))

def transform_links(task_records, task_objs):
    cal = PLANTACalendar()
    for task_rec in task_records:

        links = fetch_data.get_links_for_task(task_rec)
        current_task_uid = task_records.index(task_rec) + 1

        for link in links:
            if link.pr_id.get_value() != link.pr_successor_id.get_value():
                info_handler.warn(msg.get_const('link_external_task').format(
                    task_rec.task_id.get_value()))
                continue
            succ_rec = find_task_rec_by_planta_id(task_records, 
                link.task_successor_id.get_value())
            #task_records.index(task)
            pos = -1
            if succ_rec is None:
                info_handler.warn(msg.get_const('invalid_task_link').format(
                    task_rec.task_id.get_value(), link.task_successor_id.get_value()))
                continue
            else:
                pos = task_records.index(succ_rec)

            link_obj = transform_link(link, current_task_uid)

            #link lag:
            link_lag = calculate_link_lag(link, task_rec)
            link_lag = calc_correction_lag(task_rec, succ_rec, link.link_type.get_value(), link_lag, cal)
            add_link_lag(link_obj, link_lag)

            task_objs[pos].add_PredecessorLink(link_obj)
                
def calc_outlinelevel(psp_code):
    code_list = psp_code.split('.')
    return len(code_list)

def add_task_id_reference(task_rec, task_obj):
    from .. import mspdi_extended_fields
    mspdi_extended_fields.add_extended_field(task_obj,
        ('TaskID', task_rec.task_id.get_value()))
    mspdi_extended_fields.add_extended_field(task_obj,
        ('MSPID', task_obj.get_UID()))

def add_split_period(project_obj, task_obj, task_rec):
    assignments = None

    if project_obj.get_Assignments() == None:
        assignments = Assignments()
        project_obj.set_Assignments(assignments)
    else:
        assignments = project_obj.get_Assignments()

    actual_start = format.start_date_to_msp_date(task_rec.actual_start.get_value())

    ta_dup = task_rec.ta_dup.get_value()
    last_time_entry = format.finish_date_to_msp_date(ta_dup)
    day_after_last_time_entry = format.start_date_to_msp_date(ta_dup + 1)

    calc_start = format.start_date_to_msp_date(task_rec.calc_start.get_value())
    calc_end = format.finish_date_to_msp_date(task_rec.calc_end.get_value())
    day_before_calc_start = format.finish_date_to_msp_date(task_rec.calc_start.get_value()-1)

    if task_rec.duration_act.get_value() > 0 and task_rec.duration_rem.get_value() > 0:
        percent_completed = str(int(round(task_rec.duration_act.get_value() /
                        (task_rec.duration_act.get_value() +
                         task_rec.duration_rem.get_value()) * 100)))
    else:
        percent_completed = '0'

    task_obj.set_PercentComplete(percent_completed)

    assignment = Assignment(UID=task_obj.get_UID(), ResourceUID='-65535',
                            TaskUID=task_obj.get_UID(), PercentWorkComplete=percent_completed)
    assignment.set_ActualWork(format.days_to_msp_hours(task_rec.duration_act.get_value()))
    assignment.set_Work(format.days_to_msp_hours(
        task_rec.duration_act.get_value() + task_rec.duration_rem.get_value()))
    assignment.set_RemainingWork(format.days_to_msp_hours(task_rec.duration_rem.get_value()))
    assignment.set_RegularWork(assignment.get_Work())

    assignment.set_ActualStart(actual_start)
    assignment.set_Finish(calc_end)
    assignment.set_Start(actual_start)
    assignment.set_Stop(last_time_entry)
    assignment.set_Resume(calc_start)
    
    task_obj.set_Stop(last_time_entry)
    task_obj.set_Resume(calc_start)

    uid = assignment.get_UID()

    task_start = TimephasedDataType(UID=uid, Start=actual_start,
        Finish=last_time_entry, Value=percent_completed, Unit='2', Type='11')

    assgmnt_start = TimephasedDataType(UID=uid, Start=actual_start,
        Finish=last_time_entry, Value=task_obj.get_ActualDuration(), Unit='2', Type='2')

    assgmnt_finish = TimephasedDataType(UID=uid, Start=day_before_calc_start, Finish=calc_end,
        Value=task_obj.get_RemainingDuration(), Unit='2', Type='1')

    task_obj.add_TimephasedData(task_start)
    assignment.add_TimephasedData(assgmnt_start)
    assignment.add_TimephasedData(assgmnt_finish)    
    assignments.add_Assignment(assignment)

def clear_summary_task(msp_task):
    msp_task.set_EarlyFinish(None)
    msp_task.set_LateFinish(None)
    msp_task.set_EarlyStart(None)
    msp_task.set_EarlyFinish(None)
    msp_task.set_Start(None)
    msp_task.set_Finish(None)
    msp_task.set_ConstraintDate(None)
    msp_task.set_ConstraintType(None)
    msp_task.set_Duration(None)
    
def transform_tasks(proj_rec, project_obj, export_resources):
    tasks = fetch_data.get_task_records(proj_rec)
    tasks_obj = Tasks()
    task_objects = list()

    # MS Project needs the tasks in the right order to make the WBS/Outline
    # thing work correctly.
    # The PSP-Code gives us some similar functionality: using this code,
    # we can sort by this attribute, determine the outlinelevel by the 
    # PSP-Code elements and we can then set the WBS Code to the PSP Code.
    # 
    # If we can't rely on the PSP-Code (because it's not set in every project),
    # we'll have to build this functionality ourselves.
    tasks.sort(key=lambda task_rec:task_rec.vg_psp_code.get_value())

    for task_rec in tasks:
        task_obj = transform_task(task_rec)
        position = tasks.index(task_rec) + 1

        #UID addition
        task_obj.set_UID(position)
        task_obj.set_ID(position)

        add_task_id_reference(task_rec, task_obj)

        # Milestone special case -- really needed? 
        if task_obj.get_Milestone() == 1:
            task_obj.set_Finish(task_obj.get_Start())
            task_obj.set_EarlyFinish(None)
            task_obj.set_LateFinish(None)

        #Task hierarchy
        psp_code = task_rec.vg_psp_code.get_value()
        if psp_code == '':
            info_handler.warn(msg.get_const('no_psp').format(task_rec.task_id.get_value()))
        task_obj.set_OutlineLevel(calc_outlinelevel(psp_code))
        task_obj.set_WBS(psp_code)
        task_obj.set_OutlineNumber(psp_code)

        task_obj.set_Duration(format.days_to_msp_hours(task_rec.duration_act.get_value() + \
            task_rec.duration_rem.get_value()))

        # quick hack: delete dates/constraints from summary tasks
        #TODO: find a better transformation scheme...
        if str(task_obj.get_Summary()) == '1':
            clear_summary_task(task_obj)
        elif task_rec.actual_start.get_value() > 0 and \
            task_rec.actual_end.get_value() == 0:
            if task_rec.calc_start.get_value() > task_rec.actual_start.get_value() > 0:
                actual_start = format.start_date_to_msp_date(task_rec.actual_start.get_value())
                task_obj.set_Start(actual_start)
                task_obj.set_ConstraintDate(actual_start)
                task_obj.set_EarlyStart(task_obj.get_Start())
                task_obj.set_LateStart(task_obj.get_Start())

                add_split_period(project_obj, task_obj, task_rec)

        if task_rec.ta_is_parent_ta.get_value() > 0:
            #this task represents one subproject -> set the describing Name to
            #something like "Subproject: <name of the subproject>"
            sub_proj_name = ppms.db_select("SELECT DI000690 FROM DT461 WHERE DI001003 = "
                "'{0}' AND DI001004 = '{1}'".format(task_rec.pr_id.get_value(),
                task_rec.task_id.get_value()))[0][0]
            info_handler.info(msg.get_const('represents_subproject').format(task_rec.task_id.get_value(), sub_proj_name))                                
            task_obj.set_Name(msg.get_const('subproj_prefix').format(sub_proj_name))

        #add calendar if configured:
        if task_rec.task_calendar.get_value() not in ('', '-'):
            task_cal_uid = transform_calendar.get_extra_cal(
                task_rec.task_calendar.get_value(), project_obj)
            task_obj.set_CalendarUID(task_cal_uid)

        #Add task object to the list of task objects in the project object
        task_objects.append(task_obj)

    #process links
    transform_links(tasks, task_objects)
    
    #calculate task hash (after links were added)
    task_hash.run_link_switch(task_objects)
    for task in task_objects:
        task_hash.set_task_hash(task)
    
    if export_resources == True:
        # tasks are dependant of their assigned resources regarding the
        # duration outcome, thus have to be manipulated there, too
        #process the resourcelinks and resources  
        transform_resource.process_resource_tuples(
            tasks, task_objects, project_obj)
    
    #add the list of tasks to the <Tasks> node
    tasks_obj.set_Task(task_objects)
    
    return tasks_obj 
