from ..format import *
from ..mspdi_2003_classes import *

def transform_record(planta_record, source_object, transformation_triple):
    (di_id, format_func, getter_func) = transformation_triple
    dtp_di = planta_record.get_di_by_id(di_id)
    if not dtp_di:
        return
    planta_value = format_func(getter_func(source_object))

    if not planta_value:
        return
    dtp_di.set_value(planta_value)

def transform_whole_node(planta_record, source_object, transformation_triples):
    for triple in transformation_triples:
        transform_record(planta_record, source_object, triple)

def add_record(table_id, key_value_dict, source_object, transformation_triples):
    for di_name, transformation_func, getter in transformation_triples:
        value = getter(source_object)
        key_value_dict[di_name] = transformation_func(value)
    return ppms.create_record(table_id, key_value_dict)

# The following transformation_functions will only transform data in one MSP class
# that can be mapped to the ppms equivalent without a larger context
# All transformation that relies on knowing overall structure/a bigger context
# has to be implemented in the outer recursive transformation construct
def transform_project_data(proj_record, proj_object):
    project_triples = [
        # (DI-ID,    format_func,    setter_func)
        (7802, identity, Project.get_Company),
        (1005, msp_date_to_planta_date, Project.get_CreationDate),
        (1006, msp_date_to_planta_date, Project.get_LastSaved),
        (1019, msp_date_to_planta_date, Project.get_StartDate),
        (1020, msp_date_to_planta_date, Project.get_FinishDate),
        (690, identity, Project.get_Name)
    ]
    
    transform_whole_node(proj_record, proj_object, project_triples)

def transform_calendar_data(cal_record, cal_object):
    cal_triples = [
        # (DI-ID,    format_func,    setter_func)
        (305, identity, Calendar.set_Name)
    ]
    
    transform_whole_node(cal_record, cal_object, cal_triples)

def transform_period_data(period_record, period_object):
    period_triples = [
        # (DI-ID,    format_func,    setter_func)
    ]
    transform_whole_node(period_record, period_object, period_triples)
    
def transform_task_data(project_rec, task_object):
    task_triples = [
        # (DI-ID,    format_func,    setter_func)
        ('created_on', msp_date_to_planta_date, Task.get_CreateDate), #1099, 
        ('task_name', identity, Task.get_Name), #807
        ('esd', msp_date_to_planta_date, Task.get_EarlyStart), #1115,
        ('eed', msp_date_to_planta_date, Task.get_EarlyFinish), #1116, 
        ('lsd', msp_date_to_planta_date, Task.get_LateStart), #1117, 
        ('led', msp_date_to_planta_date, Task.get_LateFinish), #1118,
        ('requested_start', msp_date_to_planta_date, Task.get_Start),
        ('requested_end', msp_date_to_planta_date, Task.get_Finish),
        ('actual_start', msp_date_to_planta_date, Task.get_ActualStart), #1121,
        ('actual_end', msp_date_to_planta_date, Task.get_ActualFinish), #1122,
        ('duration_act', msp_hours_to_planta_days, Task.get_ActualDuration), #1132, 
        ('duration_rem', msp_hours_to_planta_days, Task.get_RemainingDuration), #1133, 
        ('milestone', text_to_int, Task.get_Milestone), #6786, 
        ('summary_task', text_to_int, Task.get_Summary) #8845,
    ]
    
    return add_record(463,
        {
            'pr_id':project_rec.pr_id.get_value(),
            'parent_task':None,
            'parent_pr_id':None,
            'split':0,
            'fix':0,
            'vg_psp_code':None
        }, task_object, task_triples)

def transform_task_comparison_data(project_rec, task_object, planta_id, change_state):
    task_triples = [
        ('task_name', identity, Task.get_Name),
        ('start_date', msp_date_to_planta_date, Task.get_Start),
        ('finish_date', msp_date_to_planta_date, Task.get_Finish),
        ('duration', msp_hours_to_planta_days, Task.get_Duration),
        ('percent_complete', text_to_float, Task.get_PercentComplete),
        ('summary_task', text_to_int, Task.get_Summary),
        ('milestone', text_to_int, Task.get_Milestone),
        ('outline_code', identity, Task.get_OutlineNumber)
    ]

    msp_id = None
    project_id = project_rec.pr_id.get_value()
    if planta_id is not None and planta_id != '-':
        existing_recs = ppms.db_select("SELECT MSP_ID FROM MSPI_TASK WHERE PR_ID = "
            "'{0}' AND TASK_ID = '{1}'".format(project_rec.pr_id.get_value(), planta_id))
        if len(existing_recs) > 0:
            msp_id = int(max(existing_recs, key = lambda x: x[0])[0])

    return add_record(806,
        {
            'MSP_ID':msp_id,
            'pr_id':project_id,
            'task_id':planta_id,
            'change_state':change_state,
            'parent_task':None
        }, task_object, task_triples)

def transform_link_comparison_data(comp_succ_rec, comp_pred_rec, link):
    link_triples = [
        ('link_type', msp_link_type_to_planta_link_type, PredecessorLink.get_Type)
        ]

    return add_record(807,
        {
            'pred_MSP_ID':comp_pred_rec.MSP_ID.get_value(),
            'succ_MSP_ID':comp_succ_rec.MSP_ID.get_value(),
        }, link, link_triples)

def transform_link_data(task_succ_rec, task_pred_rec, link_object):
    link_triples = [
        # (DI-ID,    format_func,    setter_func)
        ('link_type', msp_link_type_to_planta_link_type, PredecessorLink.get_Type),
        ('time_interval', msp_link_lag_to_planta_link_lag, PredecessorLink.get_LinkLag)
    ]

    #there are other fields that need default values... have a look!
    return add_record(465,
        {
            'pr_successor_id':task_succ_rec.pr_id.get_value(),
            'task_successor_id':task_succ_rec.task_id.get_value(),
            'pr_id':task_pred_rec.pr_id.get_value(),
            'predecessor_id':task_pred_rec.task_id.get_value(),
            'ext_ta':0,
            'link_type_percent':0,
            'link_min':1
        }, link_object, link_triples)
