import io
import base64
import rpyc
from csrpc.core.exceptions import RpcCommunicationError

from csrpc.utils import access


__all__ = ['choose_remote_file',
           'open_remote_file', 'read_remote_file',
           'copy_local_file_remote', 'copy_remote_file_locally',
           'get_file', 'store_file']

BUFSZ = 2 * 1024 * 1024


def init_file_service():
    """initializes the service for the session; if that was done before, this is a noop."""
    return access.remote_service('file_handling')


class RemoteFile(object):
    """A wrapper on the server for a file on the client"""

    def __init__(self, remote_handle, remote_is_binary, remote_path):
        """Initialize the RemoteFile"""
        self.remote_handle = remote_handle
        self.remote_is_binary = remote_is_binary

        self.remote_path = remote_path

    def __str__(self):
        return '<RemoteFile: "{}">'.format(self.remote_path)

    def read(self, *args, **kwargs):
        """Read from the remote file"""
        content = self.remote_handle.read(*args, **kwargs)

        if self.remote_is_binary:
            return base64.b64decode(content)

        return content

    def write(self, *args, **kwargs):
        """Write to the remote file"""
        return self.remote_handle.write(*args, **kwargs)

    def tell(self, *args, **kwargs):
        """Tell the position in the stream on the remote file"""
        return self.remote_handle.tell(*args, **kwargs)

    def seek(self, *args, **kwargs):
        """Seek a new position in the remote file"""
        return self.remote_handle.seek(*args, **kwargs)

    def close(self, *args, **kwargs):
        """Close the file handle on the client"""
        return self.remote_handle.close(*args, **kwargs)

    def __bool__(self):
        return bool(self.remote_handle)

    def __iter__(self):
        return self.remote_handle.__iter__()

    def __next__(self):
        content = next(self.remote_handle)

        if self.remote_is_binary:
            return base64.b64decode(content)

        return content

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        return self.remote_handle.__exit__(*args, **kwargs)


def choose_remote_file(save_file=False, title=None, filter=None, suggested_filename=None):
    """Opens a windows file choosing dialog on the client.

    Parameters:
        save_file             controls Dialog behavior w.r.t. reading/writing files.
                              For example choosing a new file name is only possible with "save_file=True")
        title                 specifies the dialog title
        filter                specifies the file filter; e.g.
                              "filter='XML Files (*.xml)|*.xml'" limits choosing files to xml files
        suggested_filename    optional file name suggestion

    Returns: path to the chosen file or None if the dialog was cancelled
    """
    file_svc = init_file_service()

    async_choose_file = rpyc.async(file_svc.choose_file)
    async_result = async_choose_file(save_file, title=title,
                                     filter=filter, file_name=suggested_filename)
    result = async_result.value

    return result


def open_remote_file(remote_path, mode, encoding=None):
    """Opens a remote file

    Use this method instead of calling open/open_raw directly
    on the service, as the class returned here is much
    nicer to work with.

    Returns a RemoteFile instance or raises a remote exception
    when the file could not be opened.
    """
    file_svc = init_file_service()

    if 'b' in mode:
        remote_handle = file_svc.open_raw(remote_path, mode)
        remote_is_binary = True
    else:
        remote_handle = file_svc.open(filename=remote_path, mode=mode, encoding=encoding)
        remote_is_binary = False

    if remote_handle:
        return RemoteFile(remote_handle=remote_handle, remote_is_binary=remote_is_binary, remote_path=remote_path)

    return None


def read_remote_file(remote_path, encoding='utf-8'):
    """Read a complete file located on the Client

    Returns a file-like object (io.TextIOWrapper)
    """
    byte_buffer = io.BytesIO()
    copy_remote_file_locally(remote_path=remote_path, local_file=byte_buffer)

    text_wrapper = io.TextIOWrapper(buffer=byte_buffer, encoding=encoding)
    # Seek back to the start so it looks like a freshly opened file
    # to the user
    text_wrapper.seek(0)

    return text_wrapper


def copy_remote_file_locally(remote_path, local_file):
    """Copies the content of a remote file into a local file

    remote_path     The path to a file on the client
    local_file      A file like object that should be written to.
                    Note that this function will write bytes to the file,
                    so if you're getting a file descriptor from
                    open() then you need to open your file with the mode 'wb'

    Returns the amount of bytes written
    """
    bytes_written = 0
    with open_remote_file(remote_path=remote_path, mode='rb') as remote_file:
        buf = remote_file.read(BUFSZ)

        while buf:
            bytes_written = local_file.write(buf)
            buf = remote_file.read(BUFSZ)

    return bytes_written


def copy_local_file_remote(local_file, remote_path):
    """Copies the content of a local file into a remote file

    local_file      A file like object that should be read from.
                    Note that this function will write bytes to the file,
                    so if you're getting a file descriptor from
                    open() then you need to open your file with the
                    mode 'rb'
    remote_path     The path to a file on the client

    Returns the amount of bytes written
    """
    bytes_written_total = 0
    with open_remote_file(remote_path=remote_path, mode='wb') as remote_file:
        buf = local_file.read(BUFSZ)

        while buf:
            bytes_written = remote_file.write(buf)
            buf = local_file.read(BUFSZ)

            if bytes_written is None:
                bytes_written = 0

            bytes_written_total += bytes_written

    return bytes_written_total


def get_file(title, filter, encoding='utf-8'):
    """Helper method that combines choosing a file with reading it

    Parameters:
        title           The title of the dialog on the client
        filter          specifies the file filter; e.g.
                        "filter='XML Files (*.xml)|*.xml'" limits choosing files to xml files
        encoding        The encoding of the file

    This method does not support reading the binary content of a file yet!

    The open file dialog is called and then the file from the client is
    read into a file-like object.

    Returns a file-like object when a file was chosen that contains the
    entire content of the remote file. Please note that when the file on
    the client is large it might take some time to load it fully into
    memory.

    When no file was chosen None is returned
    """
    remote_path = choose_remote_file(title=title, filter=filter)
    if remote_path:
        return read_remote_file(remote_path, encoding=encoding)

    return None


def store_file(local_file, title, filter, suggested_filename=None):
    """Helper method to quickly store a file on the client

    Parameters:
        local_file            A file-like object to read from
                              Note that this function will write bytes to the file,
                              so if you're getting a file descriptor from
                              open() then you need to open your file with the
                              mode 'rb'
        title                 The title of the dialog on the client
        filter                specifies the file filter; e.g.
                              "filter='XML Files (*.xml)|*.xml'" limits choosing files to xml files
        suggested_filename    The default filename that appears in the dialog

    The save file dialog is called and then the content of the
    local_file is written to the remote file

    Returns the amount of bytes written or None when the user
    cancels the dialog.
    """
    remote_path = choose_remote_file(True, title, filter, suggested_filename)
    if remote_path:
        return copy_local_file_remote(local_file=local_file, remote_path=remote_path)

    return None


def store_file_without_dialog(local_file, remote_path, remote_file_name):
    """Creates a file without opening a dialog and without over writing existing files

    Parameters:
        local_file            A file-like object to read from
                              Note that this function will write bytes to the file,
                              so if you're getting a file descriptor from
                              open() then you need to open your file with the
                              mode 'rb'
        remote_path           Path on the client side where the file should be stored
        remote_file_name      The file name of the new file on the client side with extension
                              e.g 'test.txt'
        """
    file_svc = init_file_service()
    remote_file_path = file_svc.create_file(remote_path, remote_file_name)
    copy_local_file_remote(local_file=local_file, remote_path=remote_file_path)